%========================================================================
% code1.m
% A very simple Navier-Stokes solver for a drop falling in a rectangular
% domain. The viscosity is taken to be a constant and a forward in time,
% centered in space discretization is used. The density is advected by a
% simple upwind scheme.
%A Front-tracking/Finite-Volume Navier-Stokes
%Solver for Direct Numerical Simulations of
%Multiphase Flows
%Grétar Tryggvason
%========================================================================
clear all
close all

%Load MD boundary conditions
M = csvread("MD_BC_rhouv.csv"); N = 44;
rhoMD = M(:,1:N);
uMD = M(:,N+1:2*N);
vMD = M(:,2*N+1:3*N);

% Domain size and physical variables
Lx=44; Ly=44; gx=0; gy=0;
rho1=20; rho2=1; 
m0=0.01; % Dynamic viscosity
usouth=0;
veast=0;
vwest=0;
time=0; % See above 

% Numerical variables
nx=44;
ny=44;
dt=1;
nstep=198; % Number of time steps to run
maxit=2000; % Max iterations for pressure-SOR
maxError=0.001; % Tolerance for pressure-SOR
beta=1.2; % Relaxation parameter for pressure-SOR
small=1e-30;

% Initialize arrays
u=zeros(nx+1,ny+2);v=zeros(nx+2,ny+1);p=zeros(nx+2,ny+2);
ut=zeros(nx+1,ny+2);vt=zeros(nx+2,ny+1);tmp1=zeros(nx+2,ny+2);
uu=zeros(nx+1,ny+1);vv=zeros(nx+1,ny+1);tmp2=zeros(nx+2,ny+2);

% Set the grid
dx=Lx/nx;dy=Ly/ny;
for i=1:nx+2
    x(i)=dx*(i-1.5);
end
for j=1:ny+2
    y(j)=dy*(j-1.5);
end

% Set density
r=zeros(nx+2,ny+2)+rho1;


%================== START TIME LOOP======================================
for is=1:nstep,is
    
    % Set boundary to MD values
    v(2:nx+1,1)=vMD(is,:); 
    vt(2:nx+1,1)=v(2:nx+1,1);
    r(2:nx+1,1)=rhoMD(is,:);
    
    % Set zerogradient at top boundary
    v(2:nx+1,ny+1)=v(2:nx+1,ny);
    p(2:nx+1,ny+2)=2*p(2:nx+1,ny+1)-p(2:nx+1,ny);
    u(1:nx+1,ny+2)=u(1:nx+1,ny+1);
    
    % Set the tangential velocity at the boundaries
    u(1:nx+1,1)=2*usouth-u(1:nx+1,2); 
    v(1,1:ny+1)=2*vwest-v(2,1:ny+1); 
    v(nx+2,1:ny+1)=2*veast-v(nx+1,1:ny+1);
    
    for i=2:nx % Temporary u-velocity
        for j=2:ny+1
            Ax=1/(4*dx)*((u(i+1,j)+u(i,j))^2-(u(i,j)+u(i-1,j))^2)+1/(4*dy)*((u(i,j+1)+u(i,j))*(v(i+1,j)+v(i,j))-(u(i,j)+u(i,j-1))*(v(i+1,j-1)+v(i,j-1))); % Eq. (2.20)
            Dx=m0*((u(i+1,j)-2*u(i,j)+u(i-1,j))/dx^2+(u(i,j+1)-2*u(i,j)+u(i,j-1))/dy^2); % Eq. (2.24)
            ut(i,j)=u(i,j)+dt*(-Ax+gx+2/(r(i+1,j)+r(i,j))*Dx); % Eq. (2.13) 
        end
    end
    
    for i=2:nx+1 % Temporary v-velocity
        for j=2:ny
            Ay=1/(4*dx)*((u(i,j)+u(i,j+1))*(v(i,j)+v(i+1,j))-(u(i-1,j+1)+u(i-1,j))*(v(i,j)+v(i-1,j)))+1/(4*dy)*((v(i,j+1)+v(i,j))^2-(v(i,j)+v(i,j-1))^2); % Eq. (2.21)
            Dy=m0*((v(i+1,j)-2*v(i,j)+v(i-1,j))/dx^2+(v(i,j+1)-2*v(i,j)+v(i,j-1))/dy^2); % Eq. (2.25)
            vt(i,j)=v(i,j)+dt*(-Ay+gy+2/(r(i,j+1)+r(i,j))*Dy); % Eq. (2.14) 
        end
    end
    
    % Compute source term and coefficient for pressure
    rt=r; % rt is same as rho, but with ghost values set to lrg
    lrg=1000; % Sets a large value of density to be used for ghost cells, see comment at the bottom of page 21
    %rt(1:nx+2,1)=lrg;
    rt(1:nx+2,ny+2)=rho1; % Modification for the outflow boundary
    rt(1,1:ny+2)=lrg;
    rt(nx+2,1:ny+2)=lrg;
    
    for i=2:nx+1
        for j=2:ny+1           
            tmp1(i,j)=1/(2*dt)*((ut(i,j)-ut(i-1,j))/dx+(vt(i,j)-vt(i,j-1))/dy); % Stores the RHS of Eq. (2.26)
            if j==2 % If cells next to south boundary, vsouth is known
                tmp2(i,j)=1/(1/dx^2*(1/(rt(i+1,j)+rt(i,j))+1/(rt(i,j)+rt(i-1,j)))+1/dy^2*1/(rt(i,j+1)+rt(i,j)));
            else
                tmp2(i,j)=1/(1/dx^2*(1/(rt(i+1,j)+rt(i,j))+1/(rt(i,j)+rt(i-1,j)))+1/dy^2*(1/(rt(i,j+1)+rt(i,j))+1/(rt(i,j)+rt(i,j-1)))); % Stores the 1st term at the RHS of Eq. (2.27)
            end
        end
    end
    
    for it=1:maxit % Implements the SOR method
        oldArray=p; 
        for i=2:nx+1
            for j=2:ny+1
                if j==2 % If cells next to south boundary, contribution of south pressure is not needed
                    p(i,j)=(1-beta)*p(i,j)+beta*tmp2(i,j)*(1/dx^2*(p(i+1,j)/(rt(i+1,j)+rt(i,j))+p(i-1,j)/(rt(i,j)+rt(i-1,j)))+...
                        1/dy^2*(p(i,j+1)/(rt(i,j+1)+rt(i,j)))-tmp1(i,j)); 
                else                  
                    p(i,j)=(1-beta)*p(i,j)+beta*tmp2(i,j)*(1/dx^2*(p(i+1,j)/(rt(i+1,j)+rt(i,j))+p(i-1,j)/(rt(i,j)+rt(i-1,j)))+...
                        1/dy^2*(p(i,j+1)/(rt(i,j+1)+rt(i,j))+p(i,j-1)/(rt(i,j)+rt(i,j-1)))-tmp1(i,j)); % Eq. (2.27)
                end
            end
        end
        if max(max(abs(oldArray-p)))<maxError % Exits the iterative loop when the difference between p^(alpha+1)-p^(alpha)<tolerance
            break
        end
    end
    it
    
    for i=2:nx % Correct x-velocity
        for j=2:ny+1
            u(i,j)=ut(i,j)-2*dt/(r(i+1,j)+r(i,j))*(p(i+1,j)-p(i,j))/dx;
        end
    end
    
    for i=2:nx+1 % Correct y-velocity
        for j=2:ny
            v(i,j)=vt(i,j)-2*dt/(r(i,j+1)+r(i,j))*(p(i,j+1)-p(i,j))/dy;
        end
    end
    
    % Advect the density field using centred difference scheme plus
    % diffusion
    ro=r; % stores the value of r at the time n
    for i=2:nx+1
        for j=2:ny+1
             % Eq. (2.36), uses CDS
            r(i,j)=ro(i,j)-dt*dx*(u(i,j)*r(floor(i+(1-sign(u(i,j)))/2),j)-u(i-1,j)*r(floor(i-1+(1-sign(u(i-1,j)))/2),j))... % Uses UDS
                    -dt/dy*(v(i,j)*r(i,floor(j+(1-sign(v(i,j)))/2))-v(i,j-1)*r(i,floor(j-1+(1-sign(v(i,j-1)))/2)));
        end
    end
    
    % Plot the results
    time=time+dt;
    uu(1:nx+1,1:ny+1)=0.5*(u(1:nx+1,2:ny+2)+u(1:nx+1,1:ny+1)); % Interpolates u-velocity values to horizontal lines
    vv(1:nx+1,1:ny+1)=0.5*(v(2:nx+2,1:ny+1)+v(1:nx+1,1:ny+1)); % Interpolates v-velocity values to vertical lines
    for i=1:nx+1
        xh(i)=dx*(i-1);
    end
    for j=1:ny+1
        yh(j)=dy*(j-1);
    end
    hold off
    contourf(x,y,flipud(rot90(r)),50,'linestyle','none')
    colormap(bone)
    colorbar
    caxis([min([rho1 rho2]) max([rho1 rho2])])
    axis equal
    axis([0 Lx 0 Ly])
    hold on
    quiver(xh,yh,flipud(rot90(uu)),flipud(rot90(vv)),'r')
    pause(0.01)
    
end
    









